
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

abstract public class AbstractObject
{
	protected AbstractShader shader;
	protected Matrix matrix;
	protected Scene scene;

	boolean inverted = false;

	public AbstractObject(Scene _scene)
	{
		scene = _scene;
		matrix = new Matrix();

		shader = new Shader(scene);

		/*
		if ( shader.obj != null )
		{
			// This shader is already in use
			throw new IdiotException();
		}
		*/
	}
	
	public void setShader(AbstractShader shader)
	{
		this.shader = shader;
		shader.setObject(this);
	}

	public abstract Intersection[] getIntersections(Ray ray);

	public Intersection getNearestIntersection(Ray ray)
	{
		Intersection[] intersection = getIntersections(ray);

		Intersection nearest = new Intersection(this, Float.MAX_VALUE);

		// If there aren't any intersections, return Float.MAX_VALUE, so the
		// intersection gets ignored.
		if (intersection == null)
			return nearest;

		Intersection temp = null;

		for (int i = 0; i < intersection.length; i++)
		{
			temp = intersection[i];

			/*
			 * > 0 check tilfjet da vi s ikke fr skringer
			 * bag ved kameraet.
			 */
			if (temp.t < nearest.t && temp.t > 0.001f)
			{
				nearest = temp;
			}
		}
		return nearest;
	}

	public abstract Vector getSurfaceNormal(Vector intersection);

	public abstract float[] getUV(Vector intersection);

	public AbstractShader getShader()
	{
		return shader;
	}

	public void invert()
	{
		inverted = !inverted;
	}

	public void transform(Matrix m)
	{
		matrix.transform(m);
	}

	public void translate(float dx, float dy, float dz)
	{
		transform(Matrix.translateMatrix(dx, dy, dz));
	}

	public void rotate(float dx, float dy, float dz)
	{
		transform(Matrix.rotateMatrix(dx, dy, dz));
	}
}