/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import java.util.LinkedList;
import java.util.Iterator;

public class CSGDifference extends AbstractCSGObject
{
	public CSGDifference( AbstractObject _left, AbstractObject _right, Scene _scene )
	{
		super( _left, _right, _scene );
		
		_right.invert();
	}

	
	public Intersection[] getIntersections( Ray ray )
	{
		// Entries and exits are paired in the arrays - even elements are entries, odd exits
		final Intersection left[] = leftObject.getIntersections( ray );
		
		if ( left == null )
		{
			return null;
		}
		
		final Intersection right[] = rightObject.getIntersections( ray );
		
		if ( right == null )
		{
			return left;
		}
		
		// Even indices are entrypoints, odd indices are exit points
		int leftIndex = 0;
		int rightIndex = 0;
		
		// Current intersections
		Intersection entry = null;
		Intersection exit = null;
		
		final LinkedList list = new LinkedList();
		
		while ( leftIndex < left.length && rightIndex < right.length )
		{
			final boolean leftEntersFirst = (left[leftIndex + ENTRY].t < right[rightIndex + ENTRY].t);
			final boolean leftExitsFirst = (left[leftIndex + EXIT].t < right[rightIndex + EXIT].t);

			// Do the intersection pairs overlap?
			if ( ( left[leftIndex + ENTRY].t < right[rightIndex + EXIT].t ) && ( right[rightIndex + ENTRY].t < left[leftIndex + EXIT].t ) )
			{
				// Dette kan optimeres lidt, jeg gider bare ikke lige nu
				if ( leftEntersFirst )
				{
					list.add( left[leftIndex + ENTRY] );
					list.add( right[rightIndex + ENTRY] );
					
					if ( leftExitsFirst )
					{
						leftIndex += 2;
					}
					else
					{
						left[leftIndex + ENTRY] = right[rightIndex + EXIT];
					
						rightIndex += 2;
					}
				}
				else // right enters first
				{
					if ( leftExitsFirst )
					{
						// The right span covers the entire left span, don't add anything,
						// Cut off the part of the right span up until the left span's exit
						// and advance to the next left span
						right[rightIndex + ENTRY] = left[leftIndex + EXIT];
						
						leftIndex += 2;
					}
					else
					{
						// Delete the first part of the left span, which is covered by the right span
						// In other words, push the left entry forward to right's exit.
						left[leftIndex + ENTRY] = right[rightIndex + EXIT];
						
						// We've passed the right span, get the next one
						rightIndex += 2;
					}
				}
			}
			else // the intersections don't overlap
			{
				if ( leftEntersFirst )
				{
					list.add( left[leftIndex + ENTRY] );
					list.add( left[leftIndex + EXIT] );
					
					leftIndex += 2;
				}
				else
				{
					rightIndex += 2;
				}
			}
		}
		
		while ( leftIndex < left.length )
		{
			list.add( left[leftIndex + ENTRY] );
			list.add( left[leftIndex + EXIT] );
			
			leftIndex += 2;
		}

		if ( list.isEmpty() )
		{
			return null;
		}
		
		final Intersection intersections[] = new Intersection[list.size()];
		
		int q = 0;
		
		// Move the intersections from the list to an array
		Iterator it = list.iterator();
		
		while ( it.hasNext() )
		{
			intersections[q++] = (Intersection)it.next();
		}
		
		return intersections;
	}
}