/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import java.util.LinkedList;
import java.util.Iterator;

public class CSGIntersection extends AbstractCSGObject
{
	public CSGIntersection( AbstractObject _left, AbstractObject _right, Scene _scene )
	{
		super( _left, _right, _scene );
	}
	

	public Intersection[] getIntersections( Ray ray )
	{
		// Entries and exits are paired in the arrays - even elements are entries, odd exits
		final Intersection left[] = leftObject.getIntersections( ray );
		
		if ( left == null )
		{
			return null;
		}
		
		final Intersection right[] = rightObject.getIntersections( ray );
		
		if ( right == null )
		{
			return null;
		}
		
		// Even indices are entrypoints, odd indices are exit points
		int leftIndex = 0;
		int rightIndex = 0;
		
		final LinkedList list = new LinkedList();
		
		while ( leftIndex < left.length && rightIndex < right.length )
		{
			final boolean leftEntersFirst = (left[leftIndex + ENTRY].t < right[rightIndex + ENTRY].t);
			final boolean leftExitsFirst = (left[leftIndex + EXIT].t < right[rightIndex + EXIT].t);

			// Do the intersection pairs overlap?
			if ( ( left[leftIndex + ENTRY].t < right[rightIndex + EXIT].t ) && ( right[rightIndex + ENTRY].t < left[leftIndex + EXIT].t ) )
			{
				// Does the ray enter the right object after it enters the left? (get the last entry)
				if ( leftEntersFirst )
				{
					list.add( right[rightIndex + ENTRY] );
				}
				else
				{
					list.add( left[leftIndex + ENTRY] );
				}
				
				// Does the ray exit the right object before it exits the left? (get the first exit)
				if ( leftExitsFirst )
				{
					list.add( left[leftIndex + EXIT] );
					leftIndex += 2;
				}
				else
				{
					list.add( right[rightIndex + EXIT] );
					rightIndex += 2;
				}
			}
			else // the intersections don't overlap
			{
				// Advance the object with the closest entry point
				if ( leftEntersFirst )
				{
					leftIndex += 2;
				}
				else
				{
					rightIndex += 2;
				}
			}
		}
		
		if ( list.isEmpty() )
		{
			return null;
		}
	
		final Intersection intersections[] = new Intersection[list.size()];
		
		int q = 0;
		
		// Move the intersections from the list to an array
		Iterator it = list.iterator();
		
		while ( it.hasNext() )
		{
			intersections[q++] = (Intersection)it.next();
		}
		
		return intersections;
	}
	
	public static void main( String args[] )
	{
		Scene scene = null;
		Shader shader;
		Matrix matrix;
		AbstractObject left, right, obj;
		
		left = new Sphere(0.5f,scene);
		left.translate(5.0f, 0.0f, 0.0f);
		shader = new Shader( scene );
		shader.setColor( Color.YELLOW, 1 );
		left.setShader( shader );
		
		right = new Plane( scene );
		right.rotate(0.0f, 0.0f, 0.0f);
		right.translate(5.0f, 0.0f, -2f);
		shader = new Shader( scene );
		shader.setColor( Color.RED, 1 );
		right.setShader( shader );
		
		obj = new CSGIntersection( right, left, scene );
		
		Ray ray = new Ray( new Vector(0,0,0), new Vector(1,0,0), 1,1);
		
		Intersection intersections[] = obj.getIntersections( ray );
		
		if ( intersections != null)
			for ( int i = 0; i < intersections.length; i++ )
			{
				System.out.println( "Intersection " + i + ": " + intersections[i].t + "(" + intersections[i].object.getClass() + ")" );
			}
	}
}
