
/*
 * (c) Copyright 2001 MyCorporation.
 * All Rights Reserved.
 */
package dk.ruc.vbr;

/**
 * @version 	1.0
 * @author
 */
public class Color
{
	final static public Color RED = new Color(1.0f, 0.0f, 0.0f);
	final static public Color YELLOW = new Color(1.0f, 1.0f, 0.0f);
	final static public Color BLUE = new Color(0.0f, 0.0f, 1.0f);
	final static public Color GREEN = new Color(0.0f, 1.0f, 0.0f);
	final static public Color PURPLE = new Color(1.0f, 0.0f, 1.0f);
	final static public Color ORANGE = new Color(1.0f, 1.0f, 0.0f);
	final static public Color LIGHT_GRAY = new Color(0.8f, 0.8f, 0.8f);
	final static public Color DARK_GRAY = new Color(0.2f, 0.2f, 0.2f);
	final static public Color BLACK = new Color(0.0f, 0.0f, 0.0f);
	final static public Color WHITE = new Color(1.0f, 1.0f, 1.0f);
	
	float r = 0.0f;
	float g = 0.0f;
	float b = 0.0f;
	
	public Color( float _r, float _g, float _b )
	{
		r = _r;
		g = _g;
		b = _b;
	}
	
	public Color( int argb )
	{
		r = ((argb >> 16) & 0xFF)/255.0f;
		g = ((argb >> 8) & 0xFF)/255.0f;
		b = ((argb >> 0) & 0xFF)/255.0f;
	}
	
	public Color add(Color color)
	{
		return new Color( this.r + color.r, this.g + color.g, this.b + color.b );
	}
	public Color add(float r, float g, float b)
	{
		return new Color( this.r + r, this.g + g, this.b + b );
	}	

    public Color scale(float amount)
	{
		return new Color( r * amount, g * amount, b * amount );
	}

	/* Temp hack s der ikke skal fls for meget kode ud */
	public int getRGB()
	{
		int ir = Math.round( Math.min( r, 1.0f ) * 255.0f ) & 0xFF;
		int ig = Math.round( Math.min( g, 1.0f ) * 255.0f ) & 0xFF;
		int ib = Math.round( Math.min( b, 1.0f ) * 255.0f ) & 0xFF;
		
		return ir << 16 | ig << 8 | ib;
	}

    public Color filter(Color color)
	{
		return new Color( this.r * color.r, this.g * color.g, this.b * color.b );
	}
	
	/**
	 * Blend the color with another. Mainly used for blending textures.
	 */
	public Color blend( Color other, float amount )
	{
		final Color temp = new Color(0,0,0);
		final float localAmount = 1.0f - amount;
		
		temp.r = other.r * amount - this.r * localAmount;
		temp.g = other.g * amount - this.g * localAmount;
		temp.b = other.b * amount - this.b * localAmount;
		
		return temp;
	}
    public Color attenuate( float distance )
	{
		float amount = 1.0f / ( 4.0f * (float)Math.PI * distance * distance );
		//float amount = 1/distance;
		//float amount = 1;
		return scale(amount); 
	}
	public String toString()
	{
		return "(" + r + ", " + g + ", " + b + ")";
	}
}
