
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

public class Cylinder extends AbstractObject 
{
	protected Vector start; //x, y, z in the matrix class
	protected Vector direction; //Defines the direction of the cylinder
	protected Vector circleDirection1; //Used to find intersection with circle ends of cylinder
	protected Vector circleDirection2; //Same as above
	protected Vector originSubStart;  //origin of intersection ray
	protected float length;
	protected float radius;
	protected float t;
	
	private float scaling = -0.0001f;
	
	public Cylinder( float radius, float length, Scene _scene )
	{
		super( _scene );
		this.length = length;
		this.radius = radius;

		start = matrix.getOrigin();
		direction = matrix.getUp();
	}
	
	public void transform( Matrix m )
	{
		super.transform( m );

		start = matrix.getOrigin();
		direction = matrix.getUp();
	}
		
	public Intersection[] getIntersections(Ray ray)
	{
		Intersection one;
		Intersection two;
		
		float t1;
		float t2;
		
		originSubStart = ray.origin.sub(start);
		
		float a = 1 - ray.direction.dot(direction)*ray.direction.dot(direction);
		float b = 2*(ray.direction.dot(originSubStart) - ray.direction.dot(direction)*originSubStart.dot(direction));
		float c = originSubStart.dot(originSubStart)-originSubStart.dot(direction)*originSubStart.dot(direction) - radius * radius;
		
		float discriminantSquared = b*b-4*a*c;
		
		if (discriminantSquared < 0)
			return null;
		
		if (a < 0.00001f && -0.00001f < a) // Same as the big describsion below. This ray is parallel with the cylinder
		{
			
			if (radius >= direction.cross(originSubStart).length())
			{
				final float denominator = direction.dot( ray.direction );	
					if (denominator == 0)
						return null; //Should never be returned. If plane and ray are parallel the ray should hit cylinder
						
				t1 = -(direction.dot( ray.direction )-direction.dot(start.add(direction.scale(scaling))))/denominator;
				t2 = -(direction.dot( ray.direction )-direction.dot(start.add(direction.scale(length-scaling))))/denominator;
			}
			else
			{
				return null;
			}
		}
		
		else
		{
			float discriminant = (float)Math.sqrt( discriminantSquared );
			
			t1 = (-b-discriminant)*0.5f/a;
			t2 = (-b+discriminant)*0.5f/a;
			//System.out.println("a: " + a + " t1: " + t1 + " t2: " +t2);
			
			float m1 = ray.direction.scale(t1).add(originSubStart).dot(direction);//ray.direction.dot(direction) * t1 - ray.direction.dot(originSubStart);
			float m2 = ray.direction.scale(t2).add(originSubStart).dot(direction);
			
			
	/* We will find the two intersections. If they do not hit the cylinder at the side
	 * it will hit at one or both ends. The ends can be wieved upon as two planes, with
	 * "direction" as the normal vector.
	*/
			if ((0 <= m1 && m1 <= length) || (0 <= m2 && m2 <= length))
			{
				if ((0 <= m1 && m1 <= length) && (0 <= m2 && m2 <= length)) //If not, one of the intersections is at the end of the cylinder
				{
					one = new Intersection( this, t1);
					two = new Intersection( this, t2);
				}
				
				else 
				{
					final float denominator = direction.dot( ray.direction );	
					if (denominator == 0)
						return null; //Should never be returned. If plane and ray are parallel the ray should hit cylinder
					
					Vector planeOrigin = start;
					if (0 <= m1 && m1 <= length) //If true m2 is outside
					{
						if (m2 >= length)
						{
							scaling = (length-scaling);
						}
						t2 = -(direction.dot( ray.direction )-direction.dot(start.add(direction.scale(scaling))))/denominator;
					}
					else //m1 must be outside
					{	
						
						if (m1 >= length)
						{
							scaling = length-scaling;
						}
						t1 = -(direction.dot( ray.direction )-direction.dot(start.add(direction.scale(scaling))))/denominator;
					}
				}
			}
			else
			{
				if ((m1 <= 0 && m2 >= length) || (m2 <= 0 && m1 >= length)) //If intersections on either side of cylinder both must hit the two planes
				{
					final float denominator = direction.dot( ray.direction );	
					if (denominator == 0)
						return null; //Should never be returned. If plane and ray are parallel the ray should hit cylinder
					
					t1 = -(direction.dot( ray.direction )-direction.dot(start.add(direction.scale(scaling))))/denominator;
					t2 = -(direction.dot( ray.direction)-direction.dot(start.add(direction.scale(length-scaling))))/denominator;
					
				}
				else // Both intersections is on same side of cylinder
					{
					return null;
					}
			}
			//System.out.println("m1: " + m1 + " m2: " + m2);
		}
		
		
		one = new Intersection( this, t1);
		two = new Intersection( this, t2);

		
		if ( one.t < two.t )
		{
			return new Intersection[] { one, two };
		}
		else
		{
			return new Intersection[] { two, one };
		}
		
		
	}
	
	public Vector getSurfaceNormal( Vector intersection )
	{
		float m = intersection.sub(start).dot(direction);
		if (0 <= m && m <= length)
		{
			return (intersection.sub(start.add(direction.scale(m)))).normalized();
		}	
		else if (m < 0)
		{
			return direction.scale(-1);
		}
		else
		{
			return direction;
		}
	}
	
	public float[] getUV( Vector intersection )
	{
		final Vector up = matrix.getUp();
		final Vector right = matrix.getRight();
		final Vector forward = matrix.getForward();
		
		final Vector intersectDir = intersection.sub( start ).normalized();
		float intersectionOffset = intersectDir.dot( direction );
		
		final float coords[] = new float[2];
		
		if ( 0 < intersectionOffset && intersectionOffset <= length )
		{
			float phi = (float)Math.acos( intersectDir.neg().dot( up ) );
			float theta = 0.5f * (float)Math.acos( ( forward.dot( intersectDir ) ) / (float)Math.sin(phi) ) / (float)Math.PI;
			
			if ( right.dot( intersectDir ) > 0 )
			{
				coords[0] = theta;
			}
			else
			{
				coords[0] = 1.0f - theta;
			}
			
			while ( intersectionOffset > 1 )
				intersectionOffset--;
				
			while ( intersectionOffset < 0 )
				intersectionOffset++;
			
			coords[1] = intersectionOffset;
		}
		else
		{
			coords[0] = coords[1] = 0.0f;
		}
		
		return coords;
	}
	
	public static void main (String args[])
	{
		Cylinder c = new Cylinder (1,5, null);
		c.translate(6,0,2);
		
		Ray r = new Ray (new Vector(0,0,0), (new Vector(1,0,0)).normalized(),1, 1);
		
		//c.intersects(r);
		
		Intersection[] fl = c.getIntersections(r);
		System.out.println("int 1: "+ fl[0].t + " int 2: " + fl[1].t);
	}
}
