
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import java.io.SyncFailedException;

public class DefaultCamera extends AbstractCamera
{
	protected Vector frustumTopLeft;
	protected Vector frustumDown;
	protected Vector frustumRight;
	protected Vector frustumOrigin;
	
	protected float frustumWidth = -1.0f;
	protected float frustumHeight = -1.0f;
	
	protected int outputWidth = -1;
	protected int outputHeight = -1;
	
	protected int outputPixel = -1;
	
	private int index = -1;
	
	public DefaultCamera(float FOV, Options options)
	{
		super(options);
		
		/* Hiv ud af options p et tidspunkt */
		outputWidth = options.getInt("width");
		outputHeight = options.getInt("height");
		
		float aspect = (float)outputHeight/(float)outputWidth;
		
		float horisontalFOV = FOV;
		
		// Calculate the width and height of the frustum-plane.
		float halfWidth = (float) Math.tan( 0.5f * horisontalFOV * (float)Math.PI / 180.0f );
		float halfHeight = aspect * halfWidth;
		
		frustumWidth = 2.0f * halfWidth;
		frustumHeight = 2.0f * halfHeight;
		
        System.out.println( "Width: " + outputWidth + ", height: " + outputHeight + ", aspect: " + aspect + ", half frustum width: " + frustumWidth + ", half frustum height: " + frustumHeight + ", horisontal fov: " + ((float)Math.atan( halfWidth ) * 180.0f/(float)Math.PI) + ", vertical fov: " + ((float)Math.atan(halfHeight) * 180.0f/(float)Math.PI) );
		
		// Create plane facing the negative x-axis (1,0,0)
		frustumOrigin = matrix.getOrigin();
		frustumTopLeft = matrix.transform( new Vector( 1.0f, halfWidth, halfHeight ) ).sub(frustumOrigin);
		frustumDown = matrix.getUp().neg();
		frustumRight = matrix.getRight().neg();
		
//		System.out.println( "up: " + matrix.getUp() + "(" + matrix.getUp().length() + ")\nright: " + matrix.getRight() + "(" + matrix.getRight().length() + ")" );
	}

	public void transform(Matrix m)
	{
		super.transform(m);

		// Create plane facing the negative x-axis (1,0,0)
		frustumOrigin = matrix.getOrigin();
		frustumTopLeft = matrix.transform(new Vector(1.0f, frustumWidth * 0.5f, frustumHeight * 0.5f)).sub(frustumOrigin);
		frustumDown = matrix.getUp().neg();
		frustumRight = matrix.getRight().neg();
	}

	public synchronized int getNextPixelNumber()
	{
		return ++index;
	}
	
	/**
	 * Returns an array of rays to be cast for the next pixel in the output-picture.
	 * Note that at the moment we don't check if we're inside the picture or not
	 */
	
	public synchronized Ray[] getRaysForNextPixel()
	{		
		outputPixel++; 
		return getRaysForPixel(outputPixel);
	}
	public synchronized Ray[] getRaysForPixel(int outputPixel)
	{		
		
		
		int x = outputPixel % outputWidth;
		int y = outputPixel / outputWidth;
		
		/*
		if ( y => outputHeight )
		{
			throw new IdiotException();
		}
		*/
		
		/*
		 * raydir = topleft + down * (y/outputheight) + right * (x/outputwidth) - camera
		 * raydir = raydir.normalized();
		 */

		Vector rayDir = new Vector( 0.0f, 0.0f, 0.0f );
		
//		rayDir = rayDir.add( frustumTopLeft ).add( frustumDown.scale( (float)y/(float)outputHeight ) ).add( frustumDown.scale( (float)x/(float)outputWidth ) )

		float rightscale = ( (float)x/(float)outputWidth ) * frustumWidth;
		float downscale = ( (float)y/(float)outputHeight ) * frustumHeight;

		rayDir = rayDir.add( frustumTopLeft );
		rayDir = rayDir.add( frustumDown.scale( downscale ));
		rayDir = rayDir.add( frustumRight.scale( rightscale ));
		rayDir = rayDir.normalized();
		
		Ray arrRays[] = new Ray[] { new Ray( frustumOrigin, rayDir, 1.0f, 1.0f) };
		
		return arrRays;
	}

}