
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

public class Plane extends AbstractObject 
{
	protected Vector normalVec;
	protected Matrix inverseMatrix;
	protected float d;
		
	public Plane( Scene _scene )
	{
		super( _scene );
		
		normalVec = matrix.getUp();
		inverseMatrix = matrix.inverse();
		d = normalVec.neg().dot( matrix.getOrigin() );
	}
	
	public void transform( Matrix m )
	{
		super.transform( m );
		
		normalVec = matrix.getUp();
		inverseMatrix = matrix.inverse();
		d = normalVec.neg().dot( matrix.getOrigin() );
	}
			
	public boolean intersects(Ray ray)
	{
		return (0!=normalVec.dot( ray.direction ));
	}
		
	public Intersection[] getIntersections(Ray ray)
	{
		final float denominator = normalVec.dot( ray.direction );
		if (denominator == 0)
			return null;
		
		final float t = -(normalVec.dot( ray.origin )+d)/denominator;
		
		if ( denominator < 0 )
			return new Intersection[] { new Intersection( this, t ), new Intersection( this, Float.MAX_VALUE ) };
		else
			return new Intersection[] { new Intersection( this, -Float.MAX_VALUE ), new Intersection( this, t ) };
	}
	
	public Vector getSurfaceNormal( Vector intersection )
	{
		return normalVec;
	}
		
	public float[] getUV( Vector intersection )
	{
		
		Vector trans = inverseMatrix.transform(intersection);

		//Vector projVec = intersection.sub( normalVec.scale( ( intersection.dot( normalVec ) / (normalVec.length() * normalVec.length()))));
		
		float coords[] = new float[2];
				
		coords[0] = trans.x;
		coords[1] = trans.y;
		
		return coords;
	}
	
	public static void main( String args[] )
	{	
		Plane plane = new Plane( null );
		plane.rotate(0,(float)(-Math.PI*0.5),0);
		plane.translate( -2,0,0 );

		System.out.println( "Rotated plane normal: " + plane.normalVec );
		
		Ray ray = new Ray( new Vector(0,0,0), new Vector(1,0,0), 1, 1 );
		
		Intersection intersections[] = plane.getIntersections( ray );
		
		for ( int i = 0; i < intersections.length; i++ )
			System.out.println( " " + i + ": " + intersections[i].t );
			
		System.out.println( "Random:\n" );
		
		float hej = 0;
		
		for ( int i = 0; i < 10000; i++ )
		{
			hej += Math.random();
		}
		
		System.out.println( hej );
	}
}