//her	mangler	licens	boergen

package dk.ruc.vbr;

import java.io.SyncFailedException;
import dk.ruc.vbr.photonmap.*;

public class Raytracer
{

	static Scene scene;
	AbstractOutput output;
	Options options;
	int[] bitmap;
	Color colorBitmap[];
	int index;
	boolean halt = false;
	Thread[] threads;
	int length = 0;

	// used for stats
	int totalNumberOfRays = 0;
	static int transmitted = 0;
	long time = 0;
	long photonTime = 0;

	public Raytracer(Scene scene, AbstractOutput output, Options options)
	{
		this.scene = scene;
		this.output = output;
		this.options = options;

		bitmap = output.getBitmap();
		length = bitmap.length;
		colorBitmap = new Color[bitmap.length];
	}

	public void render()
	{
		photonTime = System.currentTimeMillis();
		
		try 
		{
			PhotonTracer.emitPhotons(scene, options.getInt("photonmap"));
		}
		catch (Throwable e)
		{
			e.printStackTrace();
		}
		photonTime = System.currentTimeMillis() - photonTime;
		
		time = System.currentTimeMillis();

		int i = -1;

		while ((i = getNextPixelNumber()) < bitmap.length && halt != true)
		{
			colorBitmap[i] = Color.BLACK;

			Ray[] rays = scene.getCamera().getRaysForPixel(i);
			for (int j = 0; j < rays.length; j++)
			{
				colorBitmap[i] = colorBitmap[i].add(Render.trace(scene, rays[j]));
			}

			bitmap[i] = colorBitmap[i].scale(1f / rays.length).getRGB();
		}
		time = System.currentTimeMillis() - time;
		halt = true;

		System.out.println(
			transmitted + " " + Render.getNumberOfTraces() + " " + (System.currentTimeMillis() - time));
	}

	public static void main(String[] args)
	{
		Options options = new Options();
		options.put("width", "320");
		options.put("height", "256");
		options.put("photonmap", "1000");
		options.put("radiance", "100");
		options.put("distance", "1");
		options.put("plots", "4");
		options.put("rays", "10");

		try
		{
			scene = new SceneLoader(new java.io.File("c:\\test.scd"), options).buildScene();
			//scene = new TestSceneLoader().getScene(options);
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

		DefaultOutput output = new DefaultOutput(options);
		new Raytracer(scene, output, options).render();
	}

	protected static void increaseNumberOfTransmittedRays()
	{
		transmitted++;
	}

	synchronized int getNextPixelNumber()
	{
		index = scene.getCamera().getNextPixelNumber();
		return index;
	}

	public boolean getHalt()
	{
		return halt;
	}

	public int getIndex()
	{
		return index;
	}

	/**
	 * Gets the length.
	 * @return Returns a int
	 */
	public int getLength()
	{
		return length;
	}

	/**
	 * Gets the output.
	 * @return Returns a AbstractOutput
	 */
	public AbstractOutput getOutput()
	{
		return output;
	}
	public void stop()
	{
		halt = true;
	}


	/**
	 * Gets the time.
	 * @return Returns a long
	 */
	public long getTime()
	{
		return time;
	}

	public long getPhotonTime()
	{
		return photonTime;
	}

}