
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import java.util.*;

import dk.ruc.vbr.photonmap.PhotonMap;

public class Scene
{
	Light[] lights;
	AbstractObject[] objects;
	AbstractCamera camera;
	Color ambientColor = Color.BLACK;
	float ambientRefractionIndex = 1.0f;

	PhotonMap photonMap;

	int handle = 0;

	boolean changed = false;

	Hashtable objTable = new Hashtable();
	Hashtable lightTable = new Hashtable();

	public Scene()
	{

	}

	public void addObject(AbstractObject obj)
	{
		objTable.put(new Integer(++handle), obj);
		changed = true;
	}
	public void updateArrays()
	{
		Object[] objs = objTable.values().toArray();
		objects = new AbstractObject[objs.length];
		for (int i = 0; i < objs.length; i++)
		{
			objects[i] = (AbstractObject) objs[i];
		}
		objs = lightTable.values().toArray();
		lights = new Light[objs.length];
		for (int i = 0; i < objs.length; i++)
		{
			lights[i] = (Light) objs[i];
		}
		//objects = (vbrAbstractObject[]) objTable.values().toArray();
		changed = false;
	}
	public Light[] getLights()
	{
		if (changed)
		{
			updateArrays();
		}

		return lights;
	}
	public PhotonMap getPhotonMap()
	{
		return photonMap;
	}
	public void addLight(Light light)
	{
		lightTable.put(new Integer(++handle), light);
		changed = true;
	}
	public void addCamera(AbstractCamera camera)
	{
		this.camera = camera;
	}
	public AbstractCamera getCamera()
	{
		return camera;
	}
	public Color getIrradianceEstimate(Vector position, Vector normal, float distance, int count)
	{
		return photonMap.getIrradianceEstimate(position, normal, distance, count);
	}
	/**
	 * Gets the objects.
	 * @return Returns a AbstractObject[]
	 */
	public AbstractObject[] getObjects()
	{
		if (changed)
		{
			updateArrays();
		}
		return objects;
	}

	/**
	 * Sets the photonMap.
	 * @param photonMap The  photonMap to set
	 */
	public void setPhotonMap(PhotonMap photonMap)
	{
		this.photonMap = photonMap;
	}
	
	public void setBackgroundColor( Color value )
	{
		ambientColor = value;
	}
	
	public void setBackgroundRefractionIndex( float value )
	{
		ambientRefractionIndex = value;
	}
}