
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import electric.xml.*;
import java.awt.*;
import java.io.*;
import java.lang.reflect.*;
import java.net.URL;

import java.util.HashMap;

public class SceneLoader
{

	Document xml;
	Scene scene = new Scene();
	Options options;
	File dir;
	HashMap map = new HashMap(); 

	public SceneLoader(File file, Options options) throws Exception
	{
		dir = file.getParentFile();
		this.options = options;
		xml = new Document(file);
	}
	public Scene buildScene() throws Exception
	{

		Element root = xml.getRoot();
		Elements children = root.getElements();

		while (children.hasMoreElements())
		{
			Element element = children.next();

			String name = element.getName().toLowerCase();
			try
			{
				Class c = Class.forName(SceneLoader.class.getName());
				Method method = c.getMethod("add" + name, new Class[] { element.getClass()});
				method.invoke(this, new Object[] { element });
			}
			catch (Exception e)
			{
				e.printStackTrace();
				System.err.println(
					"Error adding element: " + System.getProperty("line.separator") + element);
			}
		}

		return scene;
	}
	public void addobject(Element element) 
	{
		System.out.println("Added object");
	}
	public void addlight(Element element)
	{
		System.out.println("Added light");

		float inten = Float.valueOf(element.getAttributeValue("int")).floatValue();

		Element rgb = element.getElement("color");
		float r = Float.valueOf(rgb.getAttributeValue("r")).floatValue();
		float g = Float.valueOf(rgb.getAttributeValue("g")).floatValue();
		float b = Float.valueOf(rgb.getAttributeValue("b")).floatValue();

		Element pos = element.getElement("position");
		float x = Float.valueOf(pos.getAttributeValue("x")).floatValue();
		float y = Float.valueOf(pos.getAttributeValue("y")).floatValue();
		float z = Float.valueOf(pos.getAttributeValue("z")).floatValue();

		Light light = new Light(new Vector(x, y, z), new Color(r, g, b).scale(inten));
		System.out.println(new Color(r, g, b).scale(inten));
		scene.addLight(light);
	}
	public void addcamera(Element element)
	{
		System.out.println("Added camera");

		int FOV = Integer.valueOf(element.getAttributeValue("fov")).intValue();
		
		AbstractCamera camera;
		
		if ( options.get( "plots" ) == "1" && options.get( "rays" ) == "1" )
		{
			camera = new DefaultCamera((FOV == 0 ? 45 : FOV), options);
		}
		else
		{
			camera = new RandomAntiAliasingCamera((FOV == 0 ? 45 : FOV), options);
		}
		
		handleRotation( camera, element.getElement("rotate") );
		handlePosition( camera, element.getElement("position") );
		
		scene.addCamera(camera);
	}
	public void addsphere(Element element)
	{
		System.out.println("Added sphere");
		
		Shader shader = new Shader(scene);
		
		final float radius = Float.valueOf(element.getAttributeValue("radius")).floatValue();

		Element surface = element.getElement("surface");
		if (surface != null)
		{
			shader = surfaceElementToShader(surface);
		}
		
		// mangler flere textures og rotation samt trans. refl. etc 

		Sphere sphere = new Sphere(radius, scene);
		
		handleRotation( sphere, element.getElement("rotate") );
		handlePosition( sphere, element.getElement("position") );
		
		sphere.setShader(shader);
		
		String name = element.getAttributeValue("name");
		if ( name != null )
			map.put( name, sphere );
		else
			scene.addObject(sphere);
	}

	public void addcylinder(Element element)
	{
		System.out.println("Added cylinder");
		
		Shader shader = new Shader(scene);
		
		float radius = Float.valueOf(element.getAttributeValue("radius")).floatValue();
		
		float length = Float.valueOf(element.getAttributeValue("length")).floatValue();
		
		Element dir = element.getElement("direction");
		float xd = Float.valueOf(dir.getAttributeValue("x")).floatValue();
		float yd = Float.valueOf(dir.getAttributeValue("y")).floatValue();
		float zd = Float.valueOf(dir.getAttributeValue("z")).floatValue();
		
		Element surface = element.getElement("surface");
		if (surface != null)
		{
			shader = surfaceElementToShader(surface);
		}
		
		// mangler flere textures og rotation samt trans. refl. etc 
		
		// og flyt dog det element ting ud i en helper klasse.. 

		Cylinder cylinder = new Cylinder(radius, length, scene);
		
		handleRotation( cylinder, element.getElement("rotate") );
		handlePosition( cylinder, element.getElement("position") );
		
		cylinder.setShader(shader);
		
		String name = element.getAttributeValue("name");
		if ( name != null )
			map.put( name, cylinder );
		else
			scene.addObject(cylinder);
	}

	public void addplane(Element element)
	{
		System.out.println("Added plane");
		
		Shader shader = new Shader(scene);
		
		Element surface = element.getElement("surface");
		if (surface != null)
		{
			shader = surfaceElementToShader(surface);
		}
		
		Plane plane = new Plane(scene);
		
		handleRotation( plane, element.getElement("rotate") );
		handlePosition( plane, element.getElement("position") );
		
		plane.setShader(shader);
		
		String name = element.getAttributeValue("name");
		if ( name != null )
			map.put( name, plane );
		else
			scene.addObject(plane);
	}
	
	public void addunion(Element element)
	{
		AbstractObject left = null, right = null;
		
		left = (AbstractObject)map.get( element.getAttributeValue( "left" ) );
		
		if ( left == null )
			return;
		
		right = (AbstractObject)map.get( element.getAttributeValue( "right" ) );
		
		if ( right == null )
			return;
		
		CSGUnion union = new CSGUnion( left, right, scene );
		
		handleRotation( union, element.getElement("rotate") );
		handlePosition( union, element.getElement("position") );

		Element surface = element.getElement("surface");
		if (surface != null)
		{
			final Shader shader = surfaceElementToShader(surface);
			union.setShader(shader);
		}
		
		String name = element.getAttributeValue("name");
		if ( name != null )
			map.put( name, union );
		else
			scene.addObject(union);
	}
	
	public void addintersection(Element element)
	{
		AbstractObject left = null, right = null;
		
		left = (AbstractObject)map.get( element.getAttributeValue( "left" ) );
		
		if ( left == null )
			return;
		
		right = (AbstractObject)map.get( element.getAttributeValue( "right" ) );
		
		if ( right == null )
			return;
		
		CSGIntersection intersection = new CSGIntersection( left, right, scene );
		
		handleRotation( intersection, element.getElement("rotate") );
		handlePosition( intersection, element.getElement("position") );

		Element surface = element.getElement("surface");
		if (surface != null)
		{
			Shader shader = surfaceElementToShader(surface);
			intersection.setShader(shader);
		}
		
		String name = element.getAttributeValue("name");
		if ( name != null )
			map.put( name, intersection );
		else
			scene.addObject(intersection);
	}
	
	public void adddifference(Element element)
	{
		AbstractObject left = null, right = null;
		
		left = (AbstractObject)map.get( element.getAttributeValue( "left" ) );
		
		if ( left == null )
			return;
		
		right = (AbstractObject)map.get( element.getAttributeValue( "right" ) );
		
		if ( right == null )
			return;
		
		CSGDifference difference = new CSGDifference( left, right, scene );
		
		handleRotation( difference, element.getElement("rotate") );
		handlePosition( difference, element.getElement("position") );

		Element surface = element.getElement("surface");
		if (surface != null)
		{
			final Shader shader = surfaceElementToShader(surface);
			difference.setShader(shader);
		}
		
		String name = element.getAttributeValue("name");
		if ( name != null )
			map.put( name, difference );
		else
			scene.addObject(difference);
	}
	
	public void addbackground( Element element )
	{
		System.out.println( "Background!" );
		String refractionString = element.getAttributeValue( "refraction" );
		
		if ( refractionString != null )
		{
			scene.setBackgroundRefractionIndex( Float.parseFloat( refractionString ) );
		}
		
		Element rgb = element.getElement("color");
		
		if ( rgb != null )
		{
			final float r = Float.valueOf(rgb.getAttributeValue("r")).floatValue();
			final float g = Float.valueOf(rgb.getAttributeValue("g")).floatValue();
			final float b = Float.valueOf(rgb.getAttributeValue("b")).floatValue();
			
			scene.setBackgroundColor( new Color( r, g, b ) );
		}
	}
	
	private Shader surfaceElementToShader(Element surface)
	{
		
		Shader shader = new Shader(scene);
		
		float samount = Float.valueOf(surface.getAttributeValue("specularamount")).floatValue();
			float sexp = Float.valueOf(surface.getAttributeValue("specularexp")).floatValue();
			
			Element rgb = surface.getElement("color");
			float r = Float.valueOf(rgb.getAttributeValue("r")).floatValue();
			float g = Float.valueOf(rgb.getAttributeValue("g")).floatValue();
			float b = Float.valueOf(rgb.getAttributeValue("b")).floatValue();

			float ramount = Float.valueOf(surface.getAttributeValue("refractionamount")).floatValue();
			float refraction = Float.valueOf(surface.getAttributeValue("refraction")).floatValue();
			float reflection = Float.valueOf(surface.getAttributeValue("reflection")).floatValue();
			
			shader.setReflectivity(reflection);
			shader.setColor(new Color(r, g, b), 1f - reflection);
			shader.setSpecular(sexp, samount);
			shader.setTranslucency(refraction, ramount);
							
			Elements textures = surface.getElements("texture");
			
			while (textures.hasMoreElements())
			{
				Element texture = textures.next();
				System.out.println("texture added");
				float offsetx = Float.valueOf(texture.getAttributeValue("offsetx")).floatValue();
				float offsety = Float.valueOf(texture.getAttributeValue("offsety")).floatValue();
				float scalex = Float.valueOf(texture.getAttributeValue("scalex")).floatValue();
				float scaley = Float.valueOf(texture.getAttributeValue("scaley")).floatValue();
				System.out.println("getting texture from "+ dir.getPath() + File.separator + texture.getAttributeValue("file"));
				Image image = Toolkit.getDefaultToolkit().createImage(dir.getPath() + File.separator + texture.getAttributeValue("file"));
				MediaTracker tracker = new java.awt.MediaTracker( new java.awt.Label() );
				tracker.addImage(image, 0);
				try
				{
					tracker.waitForAll(2000);
				}
				catch ( InterruptedException e )
				{
					System.err.println( "Texture loading interrupted." );
				}
				shader.setTexture(image);
				shader.setTextureScale(scalex, scaley);
				shader.setTextureOffset(offsetx, offsety);
			}
			Element bumpmap = surface.getElement("bumpmap");
			if (bumpmap != null) 
			{
				String value = null;
				float height = 0.1f;
				float offsetx = 0.0f;
				float offsety = 0.0f;
				float scalex = 1.0f;
				float scaley = 1.0f;
				
				value = bumpmap.getAttributeValue("height");
				
				if ( value != null )
					height = Float.valueOf( value ).floatValue();
					
				value = bumpmap.getAttributeValue("offsetx");
				
				if ( value != null )
					offsetx = Float.valueOf( value ).floatValue();
					
				value = bumpmap.getAttributeValue("offsety");
				
				if ( value != null )
					offsety = Float.valueOf( value ).floatValue();
					
				value = bumpmap.getAttributeValue("scalex");
				
				if ( value != null )
					scalex = Float.valueOf( value ).floatValue();
					
				value = bumpmap.getAttributeValue("scaley");
				
				if ( value != null )
					scaley = Float.valueOf( value ).floatValue();
				
				System.out.println("getting bump from "+dir.getPath() + File.separator + bumpmap.getAttributeValue("file"));	
				Image image = Toolkit.getDefaultToolkit().createImage(dir.getPath() + File.separator + bumpmap.getAttributeValue("file"));
				MediaTracker tracker = new java.awt.MediaTracker( new java.awt.Label() );
				tracker.addImage(image, 0);
				
				try
				{
					tracker.waitForAll(2000);
				}
				catch ( InterruptedException e )
				{
					System.err.println( "Texture loading interrupted." );
				}
				
				shader.setBumpmap(image, height);
				shader.setBumpmapScale(scalex, scaley);
				shader.setBumpmapOffset(offsetx, offsety);
			}		
			
			return shader;
	}
	
	private void handlePosition( Object obj, Element element )
	{
		if ( element != null )
		{
			float x = Float.valueOf(element.getAttributeValue("x")).floatValue();
			float y = Float.valueOf(element.getAttributeValue("y")).floatValue();
			float z = Float.valueOf(element.getAttributeValue("z")).floatValue();
			
			System.out.println( "translate (" + x + "," + y + "," + z + ")" );
			
			if ( obj instanceof AbstractObject)
			{
				((AbstractObject)obj).translate( x, y, z );
				return;
			}
			
			if ( obj instanceof AbstractCamera)
			{
				((AbstractCamera)obj).translate( x, y, z );
				return;
			}
			
			System.err.println( "Uncaught position shift!?" );
		}
	}
	
	private void handleRotation( Object obj, Element element )
	{
		if ( element != null )
		{
			final float x = Float.valueOf(element.getAttributeValue("x")).floatValue();
			final float y = Float.valueOf(element.getAttributeValue("y")).floatValue();
			final float z = Float.valueOf(element.getAttributeValue("z")).floatValue();
			
			System.out.println( "rotate (" + x + "," + y + "," + z + ")" );
			
			if ( obj instanceof AbstractObject)
			{
				((AbstractObject)obj).rotate( x, y, z );
				return;
			}
			
			if ( obj instanceof AbstractCamera)
			{
				((AbstractCamera)obj).rotate( x, y, z );
				return;
			}
			
			System.err.println( "Uncaught rotation!?" );
		}
	}
	
	public static void main(String[] arg)
	{
		try
		{
			new SceneLoader(new File("c:\\test.scd"), new Options()).buildScene();
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}
}