package dk.ruc.vbr;

public class SchlickShader extends AbstractShader
{
	protected final float fresnel;
	protected final float roughness;
	protected final float isotropy;

	private final static float shine = 5;

	public final float g, s, d;
	
	protected final int radianceEstCount;
	protected final float maxdistance;
	
	public SchlickShader(Scene _scene, Options options, float _fresnel, float _roughness, float _isotropy)
	{
		super(_scene);

		fresnel = _fresnel;
		roughness = _roughness;
		isotropy = _isotropy;

		
		radianceEstCount = options.getInt("radiance");
		maxdistance = options.getFloat("distance");
		
		
		g = 4 * roughness * (1 - roughness);

		if (roughness < 0.5)
		{
			d = 0;
			s = 1 - g;
		}
		else
		{
			d = 1 - g;
			s = 0;
		}
		System.err.println(d+g+s+" "+d+" "+g+" "+s);
	}
	public Color getDirectIllumination(
		Vector incident,
		Vector intersection,
		Vector normal,
		Vector reflection)
	{
		Light[] lights = scene.getLights();

		Color directLight = Color.BLACK;

		nextlight : for (int i = 0; i < lights.length; i++)
		{
			// Get the vector from the intersection point to the current light
			final Vector diff = lights[i].position.sub(intersection);

			final Vector light = diff.normalized();

			final float v2 = light.dot(normal);

			if (v2 <= 0)
			{
				// Light is behind the surface
				continue;
			}

			// Get the length of the above light-vector
			final float length = diff.length();

			/*
			 * Search for objects blocking the light
			 */

			Ray lightRay = new Ray(intersection, light, 1, 1);

			AbstractObject[] objs = scene.getObjects();

			// Loop through every object on the path of our ray
			for (int j = 0; j < objs.length; j++)
			{
				// Get the intersection distance
				final float s = objs[j].getNearestIntersection(lightRay).t;

				// If the distance is greater than our threshold but
				// less than the distance to the light, then the object
				// must lie between the light and the intersection point.
				// Thus we've determinded that there's an object covering
				// the light, and don't need to add the light.
				if (s > 0.001 && s < length)
					continue nextlight;
			}

			Color lightColor = lights[i].color.attenuate(length);

			Vector incident_neg = incident.neg();
			Vector tangent = new Vector(1, 0, 0); // skal vre i U's retning (fra getUV)
			Vector half = (light.add(incident_neg)).normalized();
			// Half-vektor mellem indfaldsvektor fra kamera og skyggevektor mod lys
			Vector half_ = half.add(normal.scale(-half.dot(normal)));
			// H projekteret ned p tangentplanet
			final float u = incident_neg.dot(half);
			final float t = normal.dot(half);
			final float v = incident_neg.dot(normal);
			final float w = tangent.dot(half_);
			// angiver graden af vinkelret-heden mellem H og tangentvektoren

			final float BRDF =
				S(u) * (d / (float) Math.PI + g * D(t, v, v2, w) + s * F(light, reflection));
			directLight = directLight.add(lights[i].color.scale(v2 * BRDF));
		}

		return directLight;
	}
	public Color getEmittedLight(Vector intersection, Vector normal)
	{
		//System.err.println(maxdistance+" "+radianceEstCount);
		return scene.getIrradianceEstimate(intersection, normal, maxdistance, radianceEstCount);
	}

	public Color getIntersectionColor(Ray ray, float dist)
	{
		final Vector incident = ray.direction;

		final Vector intersection = ray.getVector(dist);

		final Vector normal = getSurfaceNormal(intersection);

		final Vector reflection =
			ray.direction.sub(normal.scale(2.0f * (normal.dot(ray.direction)))).normalized();

		Color intersectionColor = Color.BLACK;

//		intersectionColor = intersectionColor.add(getEmittedLight(intersection, normal));
		//System.err.println("color"+intersectionColor);
		Color color = getDirectIllumination(incident, intersection, normal, reflection);
		if (color == Color.BLACK)
		{
		//	System.err.println(intersectionColor);
		}
		intersectionColor =
			intersectionColor.add(color);

		final float rnd = (float) Math.random();
		//System.err.println(d+g+s+" "+d+" "+g+" "+s);
		if (rnd < d)
		{
			Render.trace(scene, getDiffuseReflectedRay(intersection, ray));
			// diffuse
			//	System.err.println("diffuse");
		}
		else if (rnd < d + g)
		{
			// glossy
			//	System.err.println("glossy");
			Render.trace(scene, getGlossyReflectedRay(intersection, normal, ray));
			
		}
		else if (rnd < d + g + s)
		{
			Render.trace(scene, getSpecularReflectedRay(intersection, normal, ray));
			// specular
			//	System.err.println("specular");
		}

		return intersectionColor;
	}
	public Ray getSpecularReflectedRay(Vector intersection, Vector normal, Ray ray)
	{
		Vector v = ray.direction.sub(normal.scale(2.0f*(normal.dot(ray.direction)))).normalized();
		return new Ray(intersection, v, ray.refraction, ray.importance);
	}
	public Ray getDiffuseReflectedRay(Vector origin, Ray ray)
	{
		float a = (float) (2* Math.random() * Math.PI);
		float b = (float) Math.acos(Math.sqrt(Math.random()));
		Matrix m = new Matrix();
		m.rotate(0,a,b);
		Vector v = m.getUp().add(m.getRight()).add(m.getForward());
		return new Ray(origin, v, ray.refraction, ray.importance);
	}
	public Ray getGlossyReflectedRay(Vector origin, Vector normal, Ray ray)
	{
		float u = (float)Math.random();
		float v = (float)Math.random();
		
		float cosvr = (float)Math.pow(u,1f/(shine+1f));
		float sinvr = (float)Math.sqrt(1 - u*u);
		
		Vector O = ray.direction.neg().cross(new Vector(0,0,1));
		if (O.length() == 0)
			O = ray.direction.neg().cross(new Vector(0,1,0));
		
		// evt check for om ray direction er 0,0,1
		
		Vector P = O.cross(ray.direction.neg());
		Vector R = O.scale((float)Math.cos(2f*Math.PI*v)*sinvr);
		R = R.add(P.scale((float)Math.sin(2f*Math.PI*v)*sinvr));
		R = R.add(ray.direction.neg().scale(cosvr));
		
		Vector L = normal.scale(normal.dot(R)*2).sub(R).normalized();
		
		
		return new Ray(origin, L, ray.refraction, ray.importance);
	}
	protected float A(float w)
	{
		return (float) Math.sqrt(
			isotropy / (isotropy * isotropy - isotropy * isotropy * w * w + w * w));
	}

	protected float D(float t, float v, float v2, float w)
	{
		final float temp = G(v) * G(v2);
		return temp * Z(t) * A(w) + 1 - temp / (4 * (float) Math.PI * v * v2);
	}

	protected float G(float v)
	{
		return v / (roughness - roughness * v + v);
	}

	protected float S(float u)
	{
		final float temp = 1.0f - u;
		return fresnel + (1.0f - fresnel) * temp * temp * temp * temp * temp;
		// skaler hjlys efter indfaldsvinkel
	}

	protected float Z(float t)
	{
		final float temp = 1 - (roughness - 1) * t * t;
		return roughness / (temp * temp);
	}

	protected float F(Vector incident, Vector reflection)
	{
		//		return 1;
		//System.err.println("S="+S(incident.neg().dot(reflection)));
		return S(incident.neg().dot(reflection));
	}
	protected float F()
	{
		return 1;
	}
	public Vector getSurfaceNormal(Vector intersection)
	{

		return obj.getSurfaceNormal(intersection);

	}
}