/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

import java.awt.Image;
import java.awt.image.PixelGrabber;

public class Shader extends AbstractShader
{
	protected Color baseColor = Color.WHITE;
	protected float baseAmount = 1.0f;
	
	protected float specularAmount = 0.0f;
	protected float specularExponent = 1.0f;
	
	protected float reflectiveAmount = 0.0f;

	protected int[] texture = null;
	protected int   iTextureWidth = 1;
	protected int   iTextureHeight = 1;	
	protected float fTextureWidth = 1.0f;
	protected float fTextureHeight = 1.0f;
	protected float invTextureScaleHorisontal = 1.0f;
	protected float invTextureScaleVertical = 1.0f;
	protected float textureOffsetHorisontal = 1.0f;
	protected float textureOffsetVertical = 1.0f;
	
	protected int[] bumpmap = null;
	protected int   iBumpmapWidth = 1;
	protected int   iBumpmapHeight = 1;	
	protected float fBumpmapWidth = 1.0f;
	protected float fBumpmapHeight = 1.0f;
	protected float invBumpmapScaleHorisontal = 1.0f;
	protected float invBumpmapScaleVertical = 1.0f;
	protected float bumpmapOffsetHorisontal = 1.0f;
	protected float bumpmapOffsetVertical = 1.0f;
	
	protected float bumpmapAmount = 0.0f;
		
	protected float refractionIndex = 1.0f;
	protected float refractionAmount = 0.0f;
	
	/* ====================================================== */
	
	public Color getIntersectionColor( Ray ray, float t )
	{
		//
		Color finalColor = Color.BLACK;
		
		final Vector intersection = ray.getVector(t);
		final Vector normal = getSurfaceNormal(intersection);
		Vector reflection = null;
		
		if ( specularAmount > 0 || reflectiveAmount > 0 )
		{
			reflection = ray.direction.sub(normal.scale(2.0f*(normal.dot(ray.direction)))).normalized();
		}
		
		final Color surfaceColor = getSurfaceColor( intersection );
		
		final Color lightColor = getLighting( intersection, reflection, normal, surfaceColor );
		
		final Color reflectedColor = getReflection( intersection, reflection, ray.refraction );
		
		final Color transmittedColor = getTransmission( intersection,ray.direction, normal, ray.refraction, surfaceColor );
		
		finalColor = lightColor.scale( baseAmount );
		finalColor = finalColor.add( reflectedColor.scale( reflectiveAmount ) );
		finalColor = finalColor.add( transmittedColor.scale( refractionAmount ) );
		
		return finalColor;
	}
	
	/* ====================================================== */
	
	public void setColor( Color color, float amount )
	{
		baseColor = color;
		baseAmount = amount;
	}
	
	public void setTexture( Image img )
	{	
		fTextureWidth = iTextureWidth = img.getWidth(null);
		fTextureHeight = iTextureHeight = img.getHeight(null);
		
		texture = new int[iTextureWidth*iTextureHeight];
		
		try
		{
			PixelGrabber grabber = new PixelGrabber( img, 0, 0, iTextureWidth, iTextureHeight, texture, 0, iTextureWidth );
			grabber.grabPixels();
		}
		catch ( InterruptedException e )
		{
			System.err.println( "PixelGrabber was interrupted while grabbing the texture." );
		}
	}
	
	public void setBumpmap( Image img, float height )
	{
		bumpmapAmount = height;
		
		fBumpmapWidth = iBumpmapWidth = img.getWidth(null);
		fBumpmapHeight = iBumpmapHeight = img.getHeight(null);
		
		bumpmap = new int[iBumpmapWidth*iBumpmapHeight];
		
		try
		{
			PixelGrabber grabber = new PixelGrabber( img, 0, 0, iBumpmapWidth, iBumpmapHeight, bumpmap, 0, iBumpmapWidth );
			grabber.grabPixels();
		}
		catch ( InterruptedException e )
		{
			System.err.println( "PixelGrabber was interrupted while grabbing the bumpmap." );
		}
		
		for ( int i = 0; i < bumpmap.length; i++ )
		{
			int temp = 0;
			
			temp += bumpmap[i] & 0xFF;
			temp += (bumpmap[i] >> 8) & 0xFF;
			temp += (bumpmap[i] >> 16) & 0xFF;
			
			bumpmap[i] = temp;
		}
	}
	
	public void setSpecular( float exponent, float amount )
	{
		specularExponent = exponent;
		specularAmount = amount;
	}
	
	public void setTranslucency( float refract, float amount )
	{
		refractionIndex = refract;
		refractionAmount = amount;
	}
	
	public void setReflectivity( float amount )
	{
		reflectiveAmount = amount;
	}
	
	public void setTextureScale( float s, float t )
	{
		if ( s != 0.0f )
			invTextureScaleHorisontal = 1/s;
		
		if ( t != 0.0f )
			invTextureScaleVertical = 1/t;
	}
	
	public void setTextureOffset( float u, float v )
	{
		textureOffsetHorisontal = u;
		textureOffsetVertical = v;
	}
	
	public void setBumpmapScale( float s, float t )
	{
		if ( s != 0.0f )
			invBumpmapScaleHorisontal = 1/s;
		
		if ( t != 0.0f )
			invBumpmapScaleVertical = 1/t;
	}
	
	public void setBumpmapOffset( float u, float v )
	{
		bumpmapOffsetHorisontal = u;
		bumpmapOffsetVertical = v;
	}
	
	/* ====================================================== */
	
	protected Color getLighting( Vector intersection, Vector reflectDir, Vector normal, Color surfaceColor )
	{
		// Do we even need to calculate any lighting?
		if ( baseAmount <= 0f && specularAmount <= 0f )
		{
			return Color.BLACK;
		}
		
		Color diffuse = Color.BLACK;
		Color specular = Color.BLACK;
		
		Light[] lights = scene.getLights();

		nextlight:
		for (int i = 0; i < lights.length; i++)
		{
			// Get the vector from the intersection point to the current light
			Vector diff = lights[i].position.sub(intersection);
			
			// Get the length of the above light-vector
			float length = diff.length();

			// Turn the light-vector into a unit vector
			Vector lightDir = diff.normalized();

			float dot = normal.dot(lightDir);

			if (dot <= 0)
			{
				// Light is behind the surface
				continue;
			}
			
			/*
			 * Search for objects blocking the light
			 */

			Ray lightRay = new Ray(intersection, lightDir, 1, 1);

			AbstractObject[] objs = scene.getObjects();
			
			// Loop through every object on the path of our ray
			for ( int j = 0; j < objs.length; j++ )
			{
				if ( objs[j].inverted )
					continue;
				
				// Get the intersection distance
				final float s = objs[j].getNearestIntersection(lightRay).t;
				
				// If the distance is greater than our threshold but
				// less than the distance to the light, then the object
				// must lie between the light and the intersection point.
				// Thus we've determinded that there's an object covering
				// the light, and don't need to add the light.
				if ( s > 0.001 && s < length )
					continue nextlight;
			}
			
			Color lightColor = lights[i].color.attenuate(length);
			
			/*
			 * Diffuse lighting
			 */
			if ( baseAmount > 0f )
			{
				diffuse = diffuse.add( lightColor.scale(dot) );
			}

			/*
			 * Specular lighting
			 */
			if ( specularAmount > 0f )
			{
//				Vector reflectDir = rayDirection.sub(normal.scale(2.0f*(normal.dot(rayDirection))));
				float reflectDot = reflectDir.dot( lightDir );
				
				if ( reflectDot > 0 )
				{
					float specularIllumination = (float)Math.pow( reflectDot, specularExponent );
				
					specular = specular.add( lightColor.scale( specularIllumination * specularAmount ) );
				}
			}
		}
		
		return specular.add( diffuse.filter( surfaceColor ) );
	}
	
	protected Color getSurfaceColor( Vector intersection )
	{
		if ( texture == null )
		{
			return baseColor;
		}
		
		float uv[] = obj.getUV( intersection );
		
		uv[0] *= invTextureScaleHorisontal;
		uv[1] *= invTextureScaleVertical;
		
		float u = uv[0] * fTextureWidth + textureOffsetHorisontal;
		float v = uv[1] * fTextureHeight + textureOffsetVertical;
		
		while ( u < 0.0f )
			u += fTextureWidth;
		
		while ( v < 0.0f )
			v += fTextureHeight;
		
		while ( u >= fTextureWidth )
			u -= fTextureWidth;
		
		while ( v >= fTextureHeight )
			v -= fTextureHeight;
		
		int index = (int)u + (int)v * iTextureWidth;
		
		int alpha = (texture[index] >> 24) & 0xFF;
		
		// If the texture has an alpha mask, blend with the object's color
		if ( alpha < 255 )
		{
			float blend = alpha/255.0f;
			return baseColor.blend( new Color(texture[index]), blend );
		}
		else
		{
			return new Color( texture[index] );
		}
	}
	
	protected Vector getSurfaceNormal( Vector intersection )
	{
		if ( bumpmap == null )
		{
			if ( obj.inverted )
				return obj.getSurfaceNormal( intersection ).neg();
			else
				return obj.getSurfaceNormal( intersection );
		}
		
		// Merely replace the color member and let the TestShader do it's thing,
		float uv[] = obj.getUV( intersection );
		
		uv[0] *= invBumpmapScaleHorisontal;
		uv[1] *= invBumpmapScaleVertical;
		
		float u = uv[0] * fBumpmapWidth + bumpmapOffsetHorisontal;
		float v = uv[1] * fBumpmapHeight + bumpmapOffsetVertical;
		
		while ( u < 0.0f )
			u += fBumpmapWidth-2f;
		
		while ( v < 0.0f )
			v += fBumpmapHeight-2f;
		
		while ( u >= fBumpmapWidth-2 )
			u -= fBumpmapWidth-2;
		
		while ( v >= fBumpmapHeight-2 )
			v -= fBumpmapHeight-2;
		
		int index = (int)u + (int)v * iBumpmapWidth;
		int indexRight = index + 1;
		int indexDown = index + iBumpmapWidth;
		
		// Calculate the rightward and downward gradients on the bumpmap
		float du = (bumpmap[indexRight] - bumpmap[index])/765f + 0.5f;
		float dv = (bumpmap[indexDown] - bumpmap[index])/765f + 0.5f;
		
		du *= bumpmapAmount;
		dv *= bumpmapAmount;
		
		Vector right = new Vector( 0.0f, 1.0f, du ).normalized();
		Vector forward = new Vector( 1.0f, 0.0f, dv ).normalized();
		Vector up = forward.cross( right );
		
		Matrix matrix = new Matrix( forward, right, up );
		
		Vector normal = matrix.transform( obj.getSurfaceNormal( intersection ) );
		
		if (obj.inverted)
			return normal.neg();
		else
			return normal;
	}
	
	protected Color getReflection( Vector intersection, Vector reflection, float mediumIndex )
	{
		if ( reflectiveAmount <= 0.0f )
		{
			return Color.BLACK;
		}
		
		final Ray reflectRay = new Ray( intersection.add(reflection.scale(0.001f)), reflection, mediumIndex, reflectiveAmount);
		
		return Render.trace(scene, reflectRay);
	}
	
	protected Color getTransmission( Vector intersection, Vector incident, Vector normal, float mediumIndex, Color surfaceColor )
	{
		//intersection = intersection.add(normal.neg().scale(0.001f));
		if ( refractionAmount <= 0.0f )
		{
			return Color.BLACK;
		}
		
		final float refractionAspect = mediumIndex/refractionIndex;
		
		final float normalIncidentDot = normal.dot( incident );
		
		final float radical = 1 + refractionAspect * refractionAspect * ( normalIncidentDot * normalIncidentDot - 1);
		
		// If the radical is less than zero, total interal reflection occurs. The light keeps
		// bouncing around inside the object. Thus, return no light.
		if ( radical < 0 )
		{
			return Color.BLACK;
		}
		
		final float temp = -refractionAspect * normalIncidentDot - (float)Math.sqrt( radical );
		
		final Vector transmissionDir = incident.scale( refractionAspect ).add( normal.scale( temp ) );
		
		final Vector transmissionOrigin = intersection.add( transmissionDir.scale( 0.01f ) );
//		final Vector transmissionOrigin = intersection;
		
		final Ray transmissionRay = new Ray( transmissionOrigin, transmissionDir, mediumIndex, refractionAmount );
		System.err.println(transmissionDir+" | "+transmissionOrigin + " | " + intersection);
		return Render.trace( scene, transmissionRay ).filter( surfaceColor );
//		return null;
	}
	
	/* ====================================================== */
	
	public Shader( Scene scene )
	{
		super( scene );
	}
}
