
/* 
* ---------------------------------------------------------------------------- 
* "THE BEER-WARE LICENSE" (Revision 42): 
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote 
* this file. As long as you retain this notice you can do whatever you want 
* with this stuff. If we meet some day, and you think this stuff is worth it, 
* you can buy us a beer in return.
* ---------------------------------------------------------------------------- 
* 
*/

package dk.ruc.vbr;

public class Sphere extends AbstractObject 
{
	protected float radius;
	
	protected Vector origin;
	protected Vector up;
	protected Vector right;
	protected Vector forward;
	
	public Sphere( float _radius, Scene _scene )
	{
		super( _scene );
		
		radius = _radius;
		
		origin = matrix.getOrigin();
		forward = matrix.getForward();
		up = matrix.getUp();
		right = matrix.getRight();
	}
	
	public void transform( Matrix m )
	{
		super.transform( m );
		
		origin = matrix.getOrigin();
		forward = matrix.getForward();
		up = matrix.getUp();
		right = matrix.getRight();
	}
		
	public Intersection[] getIntersections(Ray ray)
	{
		final Vector delta = ray.origin.sub( origin );
		
		final float b = 2 * ray.direction.dot( delta );
		final float c = delta.dot( delta ) - radius*radius;
		final float discriminantSquared = b*b-4*c;
		
		if (discriminantSquared <= 0)
			return null;
		
		final float discriminant = (float)Math.sqrt( discriminantSquared );
		
		final Intersection one = new Intersection( this, (-b-discriminant)*0.5f );
		final Intersection two = new Intersection( this, (-b+discriminant)*0.5f );
		
		if ( one.t < two.t )
		{
			return new Intersection[] { one, two };
		}
		else
		{
			return new Intersection[] { two, one };
		}
	}
	
	public Vector getSurfaceNormal( Vector intersection )
	{
		return intersection.sub( origin ).normalized();
	}
	
	public float[] getUV( Vector intersection )
	{
		Vector intersectDir = intersection.sub( origin ).normalized();
		
		float coords[] = new float[2];
		
		float phi = (float)Math.acos( intersectDir.neg().dot( up ) );
		float theta = 0.5f * (float)Math.acos( ( forward.dot( intersectDir ) ) / (float)Math.sin(phi) ) / (float)Math.PI;
		
		if ( right.dot( intersectDir ) > 0 )
		{
			coords[0] = theta;
		}
		else
		{
			coords[0] = 1.0f - theta;
		}
		
		coords[1] = 1.0f - phi/(float)Math.PI;
		
		return coords;
	}
}
