/*
* ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote
* this file. As long as you retain this notice you can do whatever you want
* with this stuff. If we meet some day, and you think this stuff is worth it,
* you can buy us a beer in return.
* ----------------------------------------------------------------------------
*
*/

package dk.ruc.vbr;

public class Triangle extends AbstractObject
{

	//protected float d;

	private Vector v1, v2, v3, normal;
	private float cachedDist = Float.MAX_VALUE;

	public Triangle(Scene scene, Vector v1, Vector v2, Vector v3)
	{
		super(scene);

		this.v1 = v1;
		this.v2 = v2;
		this.v3 = v3;

		Vector vec0 = v2.sub(v1);
		Vector vec1 = v3.sub(v2);

		this.normal = vec0.cross(vec1).normalized();
		//d = -normal.x * matrix.x - normal.y * matrix.y - normal.z * matrix.z;

		//evt check for om normalen har lngde 1
	}
	
	public void transform( Matrix m )
	{
		super.transform( m );
		
		v1 = m.transform( v1 );
		v2 = m.transform( v2 );
		v3 = m.transform( v3 );
		
		normal = v2.sub( v1 ).cross( v3.sub( v1 ) ).normalized();
	}

	public boolean intersects(Ray ray)
	{
		float edgeDotiP3d;
		float D, tempD;
		int axis, nc, sh, nsh;

		int i, j;
		float uCoor[] = new float[3];
		float vCoor[] = new float[3];

		Vector vec0 = v2.sub(v1);
		Vector vec1 = v3.sub(v2);

		float vecNrmDotdirection = normal.dot(ray.direction);
		// Ray is parallel to plane

		if (vecNrmDotdirection == 0.0)
		{
			return false;
		}

		D = normal.dot(v1);
		float dist = - (normal.dot(ray.origin) - D) / vecNrmDotdirection;

		if (dist < 0.0)
		{
			return false;
		}

		// Now we know that the Ray intersects the Plane in
		Vector distance = ray.direction.scale(dist).add(ray.origin);

		// Determine where to project Triangle

		float absNrmX = Math.abs(normal.x);
		float absNrmY = Math.abs(normal.y);
		float absNrmZ = Math.abs(normal.z);

		if (absNrmX > absNrmY)
		{
			axis = 0;
		}
		else
		{
			axis = 1;
		}

		if (axis == 0)
		{
			if (absNrmX < absNrmZ)
			{
				axis = 2;
			}
		}
		else if (axis == 1)
		{
			if (absNrmY < absNrmZ)
			{
				axis = 2;
			}
		}

		switch (axis)
		{
			case 0 :
				uCoor[0] = v1.y - distance.y;
				vCoor[0] = v1.z - distance.z;
				uCoor[1] = v2.y - distance.y;
				vCoor[1] = v2.z - distance.z;
				uCoor[2] = v3.y - distance.y;
				vCoor[2] = v3.z - distance.z;
				break;

			case 1 :
				uCoor[0] = v1.x - distance.x;
				vCoor[0] = v1.z - distance.z;
				uCoor[1] = v2.x - distance.x;
				vCoor[1] = v2.z - distance.z;
				uCoor[2] = v3.x - distance.x;
				vCoor[2] = v3.z - distance.z;
				break;

			case 2 :
				uCoor[0] = v1.x - distance.x;
				vCoor[0] = v1.y - distance.y;
				uCoor[1] = v2.x - distance.x;
				vCoor[1] = v2.y - distance.y;
				uCoor[2] = v3.x - distance.x;
				vCoor[2] = v3.y - distance.y;
				break;
		}

		// initialize number of crossing, nc.

		nc = 0;

		if (vCoor[0] < 0.0)
			sh = -1;
		else
			sh = 1;

		for (i = 0; i < uCoor.length; i++)
		{
			j = i + 1;
			if (j == uCoor.length)
				j = 0;
			if (vCoor[j] < 0.0)
				nsh = -1;
			else
				nsh = 1;
			if (nsh != sh)
			{
				if ((uCoor[i] > 0.0) && (uCoor[j] > 0.0))
				{
					nc++;
				}
				else if ((uCoor[i] > 0.0) || (uCoor[j] > 0.0))
				{
					// Line might intersect U+. We need to compute intersection on U axis.
					tempD = uCoor[i] - vCoor[i] * (uCoor[j] - uCoor[i]) / (vCoor[j] - vCoor[i]);
					if (tempD > 0)
						nc++;
				}
				sh = nsh;
			}
		}

		if ((nc % 2) == 1)
		{
			return true;
		}
		else
		{

			return false;
		}

		// algo er fra http://www.acm.org/jgt/papers/MollerTrumbore97/code.html
	}

	public Intersection[] getIntersections(Ray ray)
	{
		float edgeDotiP3d;
		float D, tempD;
		int axis, nc, sh, nsh;

		int i, j;
		float uCoor[] = new float[3];
		float vCoor[] = new float[3];

		Vector vec0 = v2.sub(v1);
		Vector vec1 = v3.sub(v2);

		float vecNrmDotdirection = normal.dot(ray.direction);
		// Ray is parallel to plane

		if (vecNrmDotdirection == 0.0)
		{
			return null;
		}

		D = normal.dot(v1);
		float dist = - (normal.dot(ray.origin) - D) / vecNrmDotdirection;

		// Ray intersect behind ray's origin

		if (dist < 0.0)
		{
			//System.out.println("false" + dist);
			return null;
		}

		// Now we know that the Ray intersects the Plane in
		Vector distance = ray.direction.scale(dist).add(ray.origin);

		// Determine where to project Triangle

		float absNrmX = Math.abs(normal.x);
		float absNrmY = Math.abs(normal.y);
		float absNrmZ = Math.abs(normal.z);

		if (absNrmX > absNrmY)
		{
			axis = 0;
		}
		else
		{
			axis = 1;
		}

		if (axis == 0)
		{
			if (absNrmX < absNrmZ)
			{
				axis = 2;
			}
		}
		else if (axis == 1)
		{
			if (absNrmY < absNrmZ)
			{
				axis = 2;
			}
		}

		switch (axis)
		{
			case 0 :
				uCoor[0] = v1.y - distance.y;
				vCoor[0] = v1.z - distance.z;
				uCoor[1] = v2.y - distance.y;
				vCoor[1] = v2.z - distance.z;
				uCoor[2] = v3.y - distance.y;
				vCoor[2] = v3.z - distance.z;
				break;

			case 1 :
				uCoor[0] = v1.x - distance.x;
				vCoor[0] = v1.z - distance.z;
				uCoor[1] = v2.x - distance.x;
				vCoor[1] = v2.z - distance.z;
				uCoor[2] = v3.x - distance.x;
				vCoor[2] = v3.z - distance.z;
				break;

			case 2 :
				uCoor[0] = v1.x - distance.x;
				vCoor[0] = v1.y - distance.y;
				uCoor[1] = v2.x - distance.x;
				vCoor[1] = v2.y - distance.y;
				uCoor[2] = v3.x - distance.x;
				vCoor[2] = v3.y - distance.y;
				break;
		}

		// initialize number of crossing, nc.

		nc = 0;

		if (vCoor[0] < 0.0)
			sh = -1;
		else
			sh = 1;

		for (i = 0; i < uCoor.length; i++)
		{
			j = i + 1;
			if (j == uCoor.length)
				j = 0;
			if (vCoor[j] < 0.0)
				nsh = -1;
			else
				nsh = 1;
			if (nsh != sh)
			{
				if ((uCoor[i] > 0.0) && (uCoor[j] > 0.0))
				{
					nc++;
				}
				else if ((uCoor[i] > 0.0) || (uCoor[j] > 0.0))
				{
					// Line might intersect U+. We need to compute intersection on U axis.
					tempD = uCoor[i] - vCoor[i] * (uCoor[j] - uCoor[i]) / (vCoor[j] - vCoor[i]);
					if (tempD > 0)
						nc++;
				}
				sh = nsh;
			}
		}

		if ((nc % 2) == 1)
		{
			// calculate the distance

			dist *= ray.direction.length();
			return new Intersection[] { new Intersection(this, dist)};
		}
		else
		{
			return null;
		}

	}

	public Vector getSurfaceNormal(Vector intersection)
	{
		return normal;
	}

	public float[] getUV(Vector intersection)
	{
		return new float[] { 0f, 0f };
	}
}