/*
* ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote
* this file. As long as you retain this notice you can do whatever you want
* with this stuff. If we meet some day, and you think this stuff is worth it,
* you can buy us a beer in return.
* ----------------------------------------------------------------------------
*
*/

package dk.ruc.vbr;

public class Vector
{
    public float x, y, z;

    /**
     * No-args contructor. Initialises the vector to (0,0,0).
     */
    public Vector()
    {
        x = y = z = 0.0f;
    }

    /**
     * Constructor with arguments.
     * Sets the coords to the supplied values.
     */
    public Vector(float _x, float _y, float _z)
    {
        this();
        x = _x;
        y = _y;
        z = _z;
    }

    /**
     * Add the vector to another vector and return the sum.
     * This should have been an operator. :(
     *
     * @return the sum of the current vector and the vector supplied as an argument
     */
    public Vector add(Vector vector)
    {
        return new Vector(this.x + vector.x, this.y + vector.y, this.z + vector.z);
    }

    /**
     * Subtract a given vector from this vector and return the difference.
     *
     * @return the difference between the current vector and the vector supplied as an argument.
     */
    public Vector sub(Vector vector)
    {
        return new Vector(this.x - vector.x, this.y - vector.y, this.z - vector.z);
    }

    /**
     * Scale the vector and return the result.
     *
     * @return the vector scaled by the scalar.
     */
    public Vector scale(float scalar)
    {
        return new Vector(this.x * scalar, this.y * scalar, this.z * scalar);
    }

    /**
     * Get the length of the vector.
     *
     * @return the length of the vector.
     */
    public float length()
    {
        return (float) Math.sqrt(x * x + y * y + z * z);
    }


    public float dot(Vector b)
    {
        return (x * b.x + y * b.y + z * b.z);

    }
    public Vector cross(Vector b)
    {
        return new Vector(y*b.z-b.y*z,z*b.x-b.z*x,x*b.y-b.x*y);
    }
    /**
     * Get a normalized vector with length 1.
     *
     * @return a vector with length 1.
     */
    public Vector normalized()
    {
        return scale(1 / length());
    }
    
    public Vector neg()
    {
    	return new Vector( -x, -y, -z );
    }

    /**
     * Standard method to describe our object with a string.
     * Nothing to see here, folks. Move along.
     *
     * @return string describing the vector.
     */
    public String toString()
    {
        return "(" + x + ", " + y + ", " + z + ")";
    }
    
    public static void main( String args[] )
    {
    	Vector vec = new Vector( 1, 2, 3 );
    	
    	System.out.println( "vec: " + vec + " (length: " + vec.length() + ")" );
    	
    	System.out.println( "vec.neg(): " + vec.neg() + " (length: " + vec.neg().length() + ")" );
    	
    	System.out.println( "vec.normalized(): " + vec.normalized() + " (length: " + vec.normalized().length() + ")" );
    	
    	Vector forward = new Vector(1,0,0), right = new Vector(0,1,0);
    	
    	System.out.println( forward + " x " + right + ": " + forward.cross( right ) );
    }
}