/*
* ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote
* this file. As long as you retain this notice you can do whatever you want
* with this stuff. If we meet some day, and you think this stuff is worth it,
* you can buy us a beer in return.
* ----------------------------------------------------------------------------
*
*/

package dk.ruc.vbr.photonmap;

import java.util.Iterator;
import java.util.LinkedList;

import dk.ruc.vbr.*;

public class PhotonMap
{

	Node root;
	LinkedList[] photonLists;
	Photon[] photons;
	int stored = 0;
	int size;
	int prevscale;
	int lightsources;
	float[] boundingBoxMinimum = { Float.MAX_VALUE, Float.MAX_VALUE, Float.MAX_VALUE };
	float[] boundingBoxMaximum = { Float.MIN_VALUE, Float.MIN_VALUE, Float.MIN_VALUE };

	/**
	 * Constructor for tPhotonMap.
	 */
	public PhotonMap(int lightsources)
	{
		this.lightsources = lightsources;
		photonLists = new LinkedList[lightsources];
		for (int i = 0; i < photonLists.length; i++)
		{
			photonLists[i] = new LinkedList();
		}
	}
	public int getStored()
	{
		return stored;
	}
	public void insert(Photon photon, int index)
	{
		if (photon.position[0] < boundingBoxMinimum[0])
			boundingBoxMinimum[0] = photon.position[0];
		if (photon.position[1] < boundingBoxMinimum[1])
			boundingBoxMinimum[1] = photon.position[1];
		if (photon.position[2] < boundingBoxMinimum[2])
			boundingBoxMinimum[2] = photon.position[2];

		if (photon.position[0] > boundingBoxMaximum[0])
			boundingBoxMaximum[0] = photon.position[0];
		if (photon.position[1] > boundingBoxMaximum[1])
			boundingBoxMaximum[1] = photon.position[1];
		if (photon.position[2] > boundingBoxMaximum[2])
			boundingBoxMaximum[2] = photon.position[2];

		photonLists[index].add(photon);
		stored++;
		//photons[stored++] = photon;
		//System.out.println("stored at "+(stored-1));
	}
	public Color getIrradianceEstimate(
		Vector position,
		Vector normal,
		float maxDistance,
		int numberOfPhotons)
	{
		Color color = new Color(0.0f, 0.0f, 0.0f);
		//System.err.println(maxDistance+" "+numberOfPhotons);
		NearestPhoton np = new NearestPhoton(position, maxDistance * maxDistance, numberOfPhotons);
		//numberOfPhotons, 0, 0, numberOfPhotons + 1, numberOfPhotons + 1);

		locatePhotons(np, root);
		//System.err.println("found "+np.found);
		if (np.found < 8)
			return Color.BLACK;

		//System.err.println(position+ ", normal "+normal);
		for (int i = 1; i < np.found; i++)
		{
			Photon temp = np.result[i];
			/*if ((temp.direction.x * normal.x + temp.direction.y * normal.y + temp.direction.z * normal.z)
				> 0.0f)
			{*/
			color = color.add(temp.power);
			//		System.err.println("power"+temp.power +" "+ color);
			//	}
			//color = color.add(temp.power);
		}
		float factor = (1.0f / (float) Math.PI) / (np.distance[0] * 1);
		//float factor = 1f;
		//	System.err.println("factor "+ factor+" "+color.scale(factor));
		return color.scale(factor);
	}
	private void locatePhotons(NearestPhoton np, Node node)
	{
		if (node.isNull)
			return;
		float dist = node.photon.position[node.axis] - np.position[node.axis];
		//System.err.println("locate: "+node + " " + node.photon.position[node.axis]+" - "+np.position[node.axis]+" = "+ dist);
		if (dist < 0f)
		{
			// search right
			//	System.err.println("locate right1");
			locatePhotons(np, node.right);
			//System.err.println("done right with "+node.photon.nr+ " "+(dist * dist) +" < "+np.distance[0]);
			if (dist * dist < np.distance[0])
			{
				//		System.err.println("locate left");
				locatePhotons(np, node.left);
			}
		}
		else
		{
			// search left
			//	System.err.println("locate left1");
			locatePhotons(np, node.left);
			if (dist * dist < np.distance[0])
			{
				//		System.err.println("locate right");
				locatePhotons(np, node.right);
			}
		}

		// find square dist between np pos and node pos.
		dist = node.photon.position[0] - np.position[0];
		float idist = dist * dist;
		dist = node.photon.position[1] - np.position[1];
		idist += dist * dist;
		dist = node.photon.position[2] - np.position[2];
		idist += dist * dist;
		//System.err.println(idist+" < "+np.distance[0]);
		if (idist < np.distance[0])
		{
			//System.err.println("got one" + idist);
			for (int i = 0; i < np.result.length; i++)
			{
				//System.err.print(np.distance[i]+", ");

			}
			//System.err.println();
			if (np.found < np.max)
			{
				np.found++;
				np.distance[np.found] = idist;
				np.result[np.found] = node.photon;
			}
			else
			{
				if (!np.heaped)
				{
					//System.err.println("building heap");
					int parent, j;
					float tdist;
					int loopcount = np.found / 2;
					Photon photon;
					//System.err.println("hej;: " + loopcount);
					for (int i = loopcount; i >= 1; i--)
					{
						//System.err.println("in for " + i + " "+loopcount);
						parent = i;
						photon = np.result[i];
						tdist = np.distance[i];
						while (parent <= loopcount)
						{
							//System.err.println("in loop");
							j = parent * 2;
							//System.err.println(tdist +" >= "+ np.distance[j] +" | "+j +" "+np.found + " " +parent +" "/*+np.distance[j + 1]*/);
							//System.err.println(((j < np.found) && (np.distance[j] < np.distance[j + 1])));
							if ((j < np.found) && (np.distance[j] < np.distance[j + 1]))
								j++;
							if (tdist >= np.distance[j])
							{
								//System.err.println("break");
								break;
							}
							np.distance[parent] = np.distance[j];
							np.result[parent] = np.result[j];
							parent = j;
						}
						np.distance[parent] = tdist;
						np.result[parent] = photon;
					}
					np.heaped = true;
				}

				// reorder the heap and insert new photon
				int parent = 1;
				int j = 2;
				while (j <= np.found)
				{
					if (j < np.found && np.distance[j] < np.distance[j + 1])
						j++;
					if (idist > np.distance[j])
						break;
					np.distance[parent] = np.distance[j];
					np.result[parent] = np.result[j];
					parent = j;
					j += j;
				}
				np.result[parent] = node.photon;
				np.distance[parent] = idist;

				np.distance[0] = np.distance[1];
			}
		}

	}
	public void balance()
	{
		for (int i = 1; i < photonLists.length; i++)
		{
			System.err.println(photonLists[i].size());
			photonLists[0].addAll(photonLists[i]);
		}
		System.err.println(photonLists[0].size());
		photons = new Photon[photonLists[0].size()];
		photons = (Photon[]) photonLists[0].toArray(photons);
		size = photons.length;
		System.err.println(size + " " + stored);
		root = balance(this.root, 0, stored - 1, 0);
		//System.out.println("done "+this.root);
	}
	private Node balance(Node current, int start, int end, int level)
	{
		/*for (int i = 0; i < photons.length; i++)
		{
			System.err.println(i+" "+photons[i]);
		
		}*/
		//System.out.println((current == root));
		// calc median
		int median = (end - start) / 2;
		median += start;
		/*
				int median = 1;
				while ((4 * median) <= (end - start + 1))
					median += median;
		
				if ((3 * median) <= (end - start + 1))
				{
					median += median;
					median += start - 1;
		
				}
				else
					median = end - median;
		*/
		//System.err.println(median+" "+end +" "+start);
		// calc axis to split

		short axis = 2;
		if ((boundingBoxMaximum[0] - boundingBoxMinimum[0])
			> (boundingBoxMaximum[1] - boundingBoxMinimum[1])
			&& (boundingBoxMaximum[0] - boundingBoxMinimum[0])
				> (boundingBoxMaximum[2] - boundingBoxMinimum[2]))
			axis = 0;
		else if (
			(boundingBoxMaximum[1] - boundingBoxMinimum[1])
				> (boundingBoxMaximum[2] - boundingBoxMinimum[2]))
			axis = 1;
		//System.err.println(median+" | "+axis+"\n");	
		medianSplit(start, end, median, axis);
		/*balancedPart[index] = originalPart[median];
		balancedPart[index].splitPlane = axis;*/

		current = new Node(photons[median], axis);

		// recursive balance rest, and calc new bounding box

		//balance left
		if (median > start)
		{

			if (start < median - 1)
			{

				//current.left = child;
				float temp = boundingBoxMaximum[axis];
				boundingBoxMaximum[axis] = current.photon.position[axis];
				//		System.err.println("llevel " + level + " " + median);
				current.left = balance(current.left, start, median - 1, level + 1);
				boundingBoxMaximum[axis] = temp;
			}
			else
			{
				//System.err.println(start +" | "+median);
				current.left = new Node(photons[start], axis);
			}
		}
		//the right
		if (median < end)
		{
			if (median + 1 < end)
			{
				//current.right = child;
				float temp = boundingBoxMinimum[axis];
				boundingBoxMinimum[axis] = current.photon.position[axis];
				//System.err.println("rlevel " + level + " " + median);
				current.right = balance(current.right, median + 1, end, level + 1);
				boundingBoxMinimum[axis] = temp;
			}
			else
			{
				current.right = new Node(photons[end], axis);
			}
		}
		/*System.err.println("TRAV!:");
		traverse(current, 0, "ting");
		System.err.println();*/
		return current;
	}
	private void medianSplit(int start, int end, int median, int axis)
	{
		int left = start;
		int right = end;

		while (right > left)
		{
			//System.out.println("right is " + right);
			float v = photons[right].position[axis];

			int i = left - 1;
			int j = right;
			while (true)
			{
				while (photons[++i].position[axis] < v)
				{
					//		System.err.println(photons[i].pos[axis]+" < "+v);
				}
				while (photons[--j].position[axis] > v && j > left)
				{
					//	System.err.println(photons[j].pos[axis]+" > "+v+" && "+j+" > " +left);
				}
				if (i >= j)
					break;
				swap(photons, i, j);
			}

			swap(photons, i, right);
			if (i >= median)
				right = i - 1;
			if (i <= median)
				left = i + 1;
		}

		//System.err.println("done split");
	}
	private void swap(Photon[] photons, int i, int j)
	{
		Photon temp = photons[i];
		photons[i] = photons[j];
		photons[j] = temp;
	}
	void traverse(Node node, int d, String r)
	{
		if (node.isNull)
		{
			System.err.println(d + " " + r + " argh");
			return;
		}
		System.err.println(r + " " + node + " " + node.photon.nr);
		traverse(node.left, ++d, "left");
		traverse(node.right, ++d, "right");
	}
	public void scale(int index)
	{
		float amount = 1f / (float) photonLists[index].size();
		for (Iterator iterator = photonLists[index].iterator(); iterator.hasNext();)
		{
			Photon element = (Photon) iterator.next();
			element.power = element.power.scale(amount);
		}
	}
	public static void main(String[] args)
	{
		PhotonMap map = new PhotonMap(4);
		for (int i = 0; i < 50000; i++)
		{
			map.insert(
				new Photon(
					new Vector((float) Math.random() * 10, (float) Math.random() * 10, (float) Math.random() * 10),
					new Vector((float) Math.random() * 10, (float) Math.random() * 10, (float) Math.random() * 10),
					new Color(1, 1, 1)),
				0);
		}

		/*for (int i = 0; i < map.photons.length; i++)
		{
		
			System.err.println(map.photons[i].nr + " " + map.photons[i] + " " );
		
		}*/
		System.err.println("start");
		long tid = System.currentTimeMillis();
		map.balance();
		//System.out.println(((System.currentTimeMillis() - tid) / 1000));
		NearestPhoton np = new NearestPhoton(new Vector(2, 1, -1), 25, 4);
		/*for (int i = 0; i < map.photons.length; i++)
		{
			float dist = map.photons[i].position[0] - np.position[0];
			float idist = dist * dist;
			dist = map.photons[i].position[1] - np.position[1];
			idist += dist * dist;
			dist = map.photons[i].position[2] - np.position[2];
			idist += dist * dist;
			System.err.println(map.photons[i].nr + " " + map.photons[i] + " " + idist);
		
		}*/
		System.err.println("locates");
		//map.traverse(map.root, 0, "root");
		tid = System.currentTimeMillis();
		map.locatePhotons(np, map.root);
		System.err.println(((System.currentTimeMillis() - tid)));
		System.err.println("nearest: " + np.distance[0]);
		for (int i = 1; i < np.result.length; i++)
		{
			if (np.result[i] == null)
				System.err.println("nill");
			else
			{
				float dist = np.result[i].position[0] - np.position[0];
				float idist = dist * dist;
				dist = np.result[i].position[1] - np.position[1];
				idist += dist * dist;
				dist = np.result[i].position[2] - np.position[2];
				idist += dist * dist;
				System.err.println(np.result[i] + " " + idist);
			}
		}
	}
}