/*
* ----------------------------------------------------------------------------
* "THE BEER-WARE LICENSE" (Revision 42):
* <doktor@dyregod.dk> <mads@danquah.dk> <hartlev@ruc.dk> <tnjr@ruc.dk> wrote
* this file. As long as you retain this notice you can do whatever you want
* with this stuff. If we meet some day, and you think this stuff is worth it,
* you can buy us a beer in return.
* ----------------------------------------------------------------------------
*
*/

package dk.ruc.vbr.photonmap;

import dk.ruc.vbr.*;

public class PhotonTracer
{

	public static void trace(Scene scene, Light light, Ray ray, int index)
	{

		// find some appropriate number
		if (ray.importance < 0.001)
		{
			return;
		}

		AbstractObject[] obj = scene.getObjects();

		float t = Float.MAX_VALUE;
		//		float temp;
		AbstractObject nearestObj = null;
		Intersection nearestIntersection = null;
		Intersection temp = null;

		for (int i = 0; i < obj.length; i++)
		{
			temp = obj[i].getNearestIntersection(ray);
			/*if (obj[i] instanceof Sphere)
			{
				if (((Sphere) obj[i]).radius != 0.05f)
				{
			*/
			if (temp.t < t)
			{
				nearestObj = temp.object;
				t = temp.t;
			}
			/*		}
				}*/
		}

		if (nearestObj == null)
			return;

		AbstractShader shader = nearestObj.getShader();
		
		Vector intersection = ray.getVector(t);
		Vector normal = nearestObj.getSurfaceNormal(intersection);

		// (
		
		SchlickShader shader2 = ((SchlickShader)nearestObj.getShader());
		final float rnd = (float) Math.random();
		//System.err.println(d+g+s+" "+d+" "+g+" "+s);
		if (rnd < shader2.d)
		{
			scene.getPhotonMap().insert(
			new Photon(intersection, ray.direction, light.color.scale(1*1000f)),
			index);
			//System.err.println("diffuse");
			trace(scene, light, shader2.getDiffuseReflectedRay(intersection, normal, ray), index);
			// diffuse
			
		} 
		else if (rnd < shader2.d + shader2.g)
		{
			// glossy
			//	System.err.println("glossy");
			trace(scene, light, shader2.getGlossyReflectedRay(intersection, normal, ray), index);
			
		} 
		else if (rnd < shader2.d + shader2.g + shader2.s)
		{
			//System.err.println("specular");
			trace(scene, light, shader2.getSpecularReflectedRay(intersection, normal, ray), index);
			// specular
			//	
		}

		/*if (nearestObj instanceof Sphere && ((Sphere) nearestObj).radius != 0.05f)
		{
					Matrix matrix = new Matrix();
					matrix.translate(intersection.x, intersection.y, intersection.z);
					AbstractObject obj2 = new Sphere(0.05f, matrix, scene);
					obj2.setShader(new Shader(scene));
		
				scene.addObject(obj2);
				//	scene.updateArrays();
		}*/
		return;

	}
	public static void emitPhotons(Scene scene, int size)
	{
		Light[] lights = scene.getLights();
		scene.setPhotonMap(new PhotonMap(lights.length));
		int count = size / lights.length;
		for (int i = 0; i < lights.length; i++)
		{
			for (int j = 0; j < count; j++)
			{
				double x, y, z;

				do
				{
					x = Math.random() * 2 - 1;
					y = Math.random() * 2 - 1;
					z = Math.random() * 2 - 1;
				}
				 
				while ((x * x + y * y + z * z) > 1);

				Vector direction = new Vector((float) x, (float) y, (float) z).normalized();
				Ray ray = new Ray(lights[i].position, direction, 1, 1);

				PhotonTracer.trace(scene, lights[i], ray, i);

			}
			scene.getPhotonMap().scale(i);
		}
		System.err.println("balancing");
		scene.getPhotonMap().balance();
		scene.updateArrays();

	}
	public static void main(String[] args)
	{
	}
}